



#include <SPI.h>
#include "SdFat.h"
#include "FreeStack.h"
#include <NMEAGPS.h>




//----------------------------------------------------------------------------------------------------------------------


// SD chip select pin.
const uint8_t SD_CS_PIN = 10; //SS


// Maximum file size in blocks.
// The program creates a contiguous file with FILE_BLOCK_COUNT 512 byte blocks.
// This file is flash erased using special SD commands.  The file will be
// truncated if logging is stopped early.

const uint32_t FILE_BLOCK_COUNT = 256000;

// log file base name.  Must be six characters or less.
#define FILE_BASE_NAME "analog"


SdFat sd;

SdBaseFile binFile;

char binName[13] = FILE_BASE_NAME "00.bin";


// Buffer definitions.
//
// The logger will use SdFat's buffer plus BUFFER_BLOCK_COUNT additional
// buffers.  QUEUE_DIM must be a power of two larger than
//(BUFFER_BLOCK_COUNT + 1).
//
#if RAMEND < 0X8FF
#error Too little SRAM
//
#elif RAMEND < 0X10FF
// Use total of two 512 byte buffers.
const uint8_t BUFFER_BLOCK_COUNT = 1;
// Dimension for queues of 512 byte SD blocks.
const uint8_t QUEUE_DIM = 4;  // Must be a power of two!
//
#elif RAMEND < 0X20FF
// Use total of five 512 byte buffers.
const uint8_t BUFFER_BLOCK_COUNT = 4;
// Dimension for queues of 512 byte SD blocks.
const uint8_t QUEUE_DIM = 8;  // Must be a power of two!
//
#elif RAMEND < 0X40FF
// Use total of 13 512 byte buffers.
const uint8_t BUFFER_BLOCK_COUNT = 12;
// Dimension for queues of 512 byte SD blocks.
const uint8_t QUEUE_DIM = 16;  // Must be a power of two!
//
#else  // RAMEND
// Use total of 29 512 byte buffers.
const uint8_t BUFFER_BLOCK_COUNT = 28;
// Dimension for queues of 512 byte SD blocks.
const uint8_t QUEUE_DIM = 32;  // Must be a power of two!
#endif  // RAMEND


// Temporary log file.  Will be deleted if a reset or power failure occurs.
#define TMP_FILE_NAME "tmp_log.bin"

// Size of file base name.  Must not be larger than six.
const uint8_t BASE_NAME_SIZE = sizeof(FILE_BASE_NAME) - 1;




// max number of blocks to erase per erase call
uint32_t const ERASE_SIZE = 262144L;



const size_t DATA_DIM16 = 249; // 3 * 83
struct block16_t {
  unsigned short count;    // count of data values
  unsigned short overrun;  // count of overruns since last block
  unsigned long  seconds;
  unsigned long  microseconds;
  unsigned short data[DATA_DIM16];
  unsigned short spare;
};

typedef block16_t block_t;


block_t* emptyQueue[QUEUE_DIM];
uint8_t emptyHead;
uint8_t emptyTail;

block_t* fullQueue[QUEUE_DIM];
volatile uint8_t fullHead;  // volatile insures non-interrupt code sees changes.
uint8_t fullTail;

// queueNext assumes QUEUE_DIM is a power of two
inline uint8_t queueNext(uint8_t ht) {
  return (ht + 1) & (QUEUE_DIM -1);
}


// Pointer to current buffer.
block_t* isrBuf;

// Need new buffer if true.
bool isrBufNeeded = true;

// overrun count
uint16_t isrOver = 0;




// Digital pin to indicate an error, set to -1 if not used.
// The led blinks for fatal errors. The led goes on solid for SD write
// overrun errors and logging continues.
const int8_t ERROR_LED_PIN = 13;

//==============================================================================
// Error messages stored in flash.
#define error(msg) errorFlash(F(msg))
//------------------------------------------------------------------------------
void errorFlash(const __FlashStringHelper* msg) {
  sd.errorPrint(msg);
  fatalBlink();
}
//------------------------------------------------------------------------------
//
void fatalBlink() {
  while (true) {
    if (ERROR_LED_PIN >= 0) {
      digitalWrite(ERROR_LED_PIN, HIGH);
      delay(200);
      digitalWrite(ERROR_LED_PIN, LOW);
      delay(200);
    }
  }
}






void logData() 
{
 uint32_t bgnBlock, endBlock;

 // Allocate extra buffer space.
 block_t block[BUFFER_BLOCK_COUNT];

  // Find unused file name.
 if(BASE_NAME_SIZE > 6) 
 {
  error("FILE_BASE_NAME too long");
 }
 
 while (sd.exists(binName)) 
 {
  if (binName[BASE_NAME_SIZE + 1] != '9') 
  {
   binName[BASE_NAME_SIZE + 1]++;
  } 
  else 
  {
   binName[BASE_NAME_SIZE + 1] = '0';
   if (binName[BASE_NAME_SIZE] == '9') 
   {
    error("Can't create file name");
   }
   binName[BASE_NAME_SIZE]++;
  }
 }
  // Delete old tmp file.
 if (sd.exists(TMP_FILE_NAME)) 
 {
  Serial.println(F("Deleting tmp file"));
  if (!sd.remove(TMP_FILE_NAME)) 
  {
   error("Can't remove tmp file");
  }
 }

 // Create new file.
 Serial.println(F("Creating new file"));
 binFile.close();
 if(!binFile.createContiguous(sd.vwd(),TMP_FILE_NAME, 512 * FILE_BLOCK_COUNT)) 
 {
  error("createContiguous failed");
 }
 // Get the address of the file on the SD.
 if (!binFile.contiguousRange(&bgnBlock, &endBlock)) 
 {
  error("contiguousRange failed");
 }

  // Use SdFat's internal buffer.
 uint8_t* cache = (uint8_t*)sd.vol()->cacheClear();
 if (cache == 0) 
 {
  error("cacheClear failed");
 }

  // Flash erase all data in the file.
 Serial.println(F("Erasing all data"));
 uint32_t bgnErase = bgnBlock;
 uint32_t endErase;
 while (bgnErase < endBlock) 
 {
  endErase = bgnErase + ERASE_SIZE;
  if(endErase > endBlock) 
  {
   endErase = endBlock;
  }
  if(!sd.card()->erase(bgnErase, endErase)) 
  {
   error("erase failed");
  }
  bgnErase = endErase + 1;
 }
 // Start a multiple block write.
 if(!sd.card()->writeStart(bgnBlock,FILE_BLOCK_COUNT)) 
 {
  error("writeBegin failed");
 }
 // Write metadata.
 if(!sd.card()->writeData((uint8_t*)&block[0])) 
 {
  error("Write metadata failed");
 }

 // Initialize queues.
 emptyHead = emptyTail = 0;
 fullHead = fullTail = 0;

 // Use SdFat buffer for one block.
 emptyQueue[emptyHead] = (block_t*)cache;
 emptyHead = queueNext(emptyHead);

  // Put rest of buffers in the empty queue.
 for(uint8_t i = 0; i < BUFFER_BLOCK_COUNT; i++) 
 {
  emptyQueue[emptyHead] = &block[i];
  emptyHead = queueNext(emptyHead);
 }

  // Give SD time to prepare for big write.
  delay(1000);
 
  Serial.flush();
  delay(10);
  uint32_t bn = 1;
  uint32_t t0 = millis();
  uint32_t t1 = t0;
  uint32_t overruns = 0;
  uint32_t count = 0;
  uint32_t maxLatency = 0;

  // Start logging interrupts.
  
  isrBuf=NULL;
  isrBufNeeded = true;
  
  adcStart();
  while (1) 
  {
    if (fullHead != fullTail) {
      // Get address of block to write.
      block_t* pBlock = fullQueue[fullTail];

      // Write block to SD.
      uint32_t usec = micros();
      if (!sd.card()->writeData((uint8_t*)pBlock)) {
        error("write data failed");
      }
      usec = micros() - usec;
      t1 = millis();
      if (usec > maxLatency) {
        maxLatency = usec;
      }
      count += pBlock->count;

      // Add overruns and possibly light LED.
      if (pBlock->overrun) {
        overruns += pBlock->overrun;
        if (ERROR_LED_PIN >= 0) {
          digitalWrite(ERROR_LED_PIN, HIGH);
        }
      }
      // Move block to empty queue.
      emptyQueue[emptyHead] = pBlock;
      emptyHead = queueNext(emptyHead);
      fullTail = queueNext(fullTail);
      bn++;
      if (bn == FILE_BLOCK_COUNT) {
        // File full so stop ISR calls.
       adcStop();
        break;
      }
    }

    
    if (Serial.available()) {
      // Stop ISR calls.
      adcStop();
      if (isrBuf != 0 && isrBuf->count >= 0) { // was PIN_COUNT
        // Truncate to last complete sample.
        //isrBuf->count = PIN_COUNT*(isrBuf->count/PIN_COUNT);
        // Put buffer in full queue.
        fullQueue[fullHead] = isrBuf;
        fullHead = queueNext(fullHead);
        isrBuf = 0;
      }

      if (fullHead == fullTail) {
        break;
      }
    }
  }
  if (!sd.card()->writeStop()) {
    error("writeStop failed");
  }
  // Truncate file if recording stopped early.
  if (bn != FILE_BLOCK_COUNT) {
    Serial.println(F("Truncating file"));
    if (!binFile.truncate(512L * bn)) {
      error("Can't truncate file");
    }
  }
  if (!binFile.rename(sd.vwd(), binName)) {
    error("Can't rename file");
  }
  Serial.print(F("File renamed: "));
  Serial.println(binName);
  Serial.print(F("Max block write usec: "));
  Serial.println(maxLatency);
  Serial.print(F("Record time sec: "));
  Serial.println(0.001*(t1 - t0), 3);
  Serial.print(F("Sample count: "));
  Serial.println(count); // was /PIN_COUNT
  Serial.print(F("Samples/sec: "));
  Serial.println((1000.0)*count/(t1-t0)); // was /PIN_COUNT (next to 1000)
  Serial.print(F("Overruns: "));
  Serial.println(overruns);
  Serial.println(F("Done"));
}




//------------------------------------------------------------------------------------------------------------




#include "i2c.h"


#define ADS1015_ADDRESS                 (0x48)    // 1001 000 (ADDR = GND)

 

#define I2C_ADDRESS ADS1015_ADDRESS

    #define ADS1015_REG_POINTER_MASK        (0x03)
    #define ADS1015_REG_POINTER_CONVERT     (0x00)
    #define ADS1015_REG_POINTER_CONFIG      (0x01)
    #define ADS1015_REG_POINTER_LOWTHRESH   (0x02)
    #define ADS1015_REG_POINTER_HITHRESH    (0x03)

    
    #define ADS1015_REG_CONFIG_OS_MASK      (0x8000)
    #define ADS1015_REG_CONFIG_OS_SINGLE    (0x8000)  // Write: Set to start a single-conversion
    #define ADS1015_REG_CONFIG_OS_BUSY      (0x0000)  // Read: Bit = 0 when conversion is in progress
    #define ADS1015_REG_CONFIG_OS_NOTBUSY   (0x8000)  // Read: Bit = 1 when device is not performing a conversion

    #define ADS1015_REG_CONFIG_MUX_MASK     (0x7000)
    #define ADS1015_REG_CONFIG_MUX_DIFF_0_1 (0x0000)  // Differential P = AIN0, N = AIN1 (default)
    #define ADS1015_REG_CONFIG_MUX_DIFF_0_3 (0x1000)  // Differential P = AIN0, N = AIN3
    #define ADS1015_REG_CONFIG_MUX_DIFF_1_3 (0x2000)  // Differential P = AIN1, N = AIN3
    #define ADS1015_REG_CONFIG_MUX_DIFF_2_3 (0x3000)  // Differential P = AIN2, N = AIN3
    #define ADS1015_REG_CONFIG_MUX_SINGLE_0 (0x4000)  // Single-ended AIN0
    #define ADS1015_REG_CONFIG_MUX_SINGLE_1 (0x5000)  // Single-ended AIN1
    #define ADS1015_REG_CONFIG_MUX_SINGLE_2 (0x6000)  // Single-ended AIN2
    #define ADS1015_REG_CONFIG_MUX_SINGLE_3 (0x7000)  // Single-ended AIN3

    #define ADS1015_REG_CONFIG_PGA_MASK     (0x0E00)
    #define ADS1015_REG_CONFIG_PGA_6_144V   (0x0000)  // +/-6.144V range = Gain 2/3
    #define ADS1015_REG_CONFIG_PGA_4_096V   (0x0200)  // +/-4.096V range = Gain 1
    #define ADS1015_REG_CONFIG_PGA_2_048V   (0x0400)  // +/-2.048V range = Gain 2 (default)
    #define ADS1015_REG_CONFIG_PGA_1_024V   (0x0600)  // +/-1.024V range = Gain 4
    #define ADS1015_REG_CONFIG_PGA_0_512V   (0x0800)  // +/-0.512V range = Gain 8
    #define ADS1015_REG_CONFIG_PGA_0_256V   (0x0A00)  // +/-0.256V range = Gain 16

    #define ADS1015_REG_CONFIG_MODE_MASK    (0x0100)
    #define ADS1015_REG_CONFIG_MODE_CONTIN  (0x0000)  // Continuous conversion mode
    #define ADS1015_REG_CONFIG_MODE_SINGLE  (0x0100)  // Power-down single-shot mode (default)

    #define ADS1015_REG_CONFIG_DR_MASK      (0x00E0)  
    #define ADS1015_REG_CONFIG_DR_128SPS    (0x0000)  // 128 samples per second
    #define ADS1015_REG_CONFIG_DR_250SPS    (0x0020)  // 250 samples per second
    #define ADS1015_REG_CONFIG_DR_490SPS    (0x0040)  // 490 samples per second
    #define ADS1015_REG_CONFIG_DR_920SPS    (0x0060)  // 920 samples per second
    #define ADS1015_REG_CONFIG_DR_1600SPS   (0x0080)  // 1600 samples per second (default)
    #define ADS1015_REG_CONFIG_DR_2400SPS   (0x00A0)  // 2400 samples per second
    #define ADS1015_REG_CONFIG_DR_3300SPS   (0x00C0)  // 3300 samples per second
    #define ADS1015_REG_CONFIG_DR_6600SPS   (0x00E0)  // 3300 samples per second
  
    #define ADS1015_REG_CONFIG_CMODE_MASK   (0x0010)
    #define ADS1015_REG_CONFIG_CMODE_TRAD   (0x0000)  // Traditional comparator with hysteresis (default)
    #define ADS1015_REG_CONFIG_CMODE_WINDOW (0x0010)  // Window comparator

    #define ADS1015_REG_CONFIG_CPOL_MASK    (0x0008)
    #define ADS1015_REG_CONFIG_CPOL_ACTVLOW (0x0000)  // ALERT/RDY pin is low when active (default)
    #define ADS1015_REG_CONFIG_CPOL_ACTVHI  (0x0008)  // ALERT/RDY pin is high when active

    #define ADS1015_REG_CONFIG_CLAT_MASK    (0x0004)  // Determines if ALERT/RDY pin latches once asserted
    #define ADS1015_REG_CONFIG_CLAT_NONLAT  (0x0000)  // Non-latching comparator (default)
    #define ADS1015_REG_CONFIG_CLAT_LATCH   (0x0004)  // Latching comparator

    #define ADS1015_REG_CONFIG_CQUE_MASK    (0x0003)
    #define ADS1015_REG_CONFIG_CQUE_1CONV   (0x0000)  // Assert ALERT/RDY after one conversions
    #define ADS1015_REG_CONFIG_CQUE_2CONV   (0x0001)  // Assert ALERT/RDY after two conversions
    #define ADS1015_REG_CONFIG_CQUE_4CONV   (0x0002)  // Assert ALERT/RDY after four conversions
    #define ADS1015_REG_CONFIG_CQUE_NONE    (0x0003)  // Disable the comparator and put ALERT/RDY in high state (default)

typedef enum
{
  GAIN_TWOTHIRDS    = ADS1015_REG_CONFIG_PGA_6_144V,
  GAIN_ONE          = ADS1015_REG_CONFIG_PGA_4_096V,
  GAIN_TWO          = ADS1015_REG_CONFIG_PGA_2_048V,
  GAIN_FOUR         = ADS1015_REG_CONFIG_PGA_1_024V,
  GAIN_EIGHT        = ADS1015_REG_CONFIG_PGA_0_512V,
  GAIN_SIXTEEN      = ADS1015_REG_CONFIG_PGA_0_256V
} adsGain_t;





void i2c_writeregister(uint8_t i2cAddress, uint8_t reg, uint16_t value)
{
 i2c_txn_t *tn;
 char       tnbuff[sizeof(i2c_txn_t)+sizeof(i2c_op)];
 char       buff[4];

 tn=(i2c_txn_t *)tnbuff;

 i2c_txn_init(tn, 1);

 buff[0]=reg;
 buff[1]=value>>8;
 buff[2]=value & 0xFF;

 
 i2c_op_init_wr(&tn->ops[0],I2C_ADDRESS,(uint8_t*)buff,3);

 i2c_post(tn);
 while (!(tn->flags & I2C_TXN_DONE)) 
 {
  /* You can do whatever you want here. */
  yield();
 }

 if (tn->flags & I2C_TXN_ERR) 
 {
  /*
   * Some error occured...
   * 
   */

   Serial.println(F("write some error occured"));
 } 
 else 
 {
  /*
   * The char[] `world` now contains the response to the read request.
   */

   Serial.println(F("OK"));
 }
}





#define DATAREADYQUEUE 3

static volatile uint8_t channel=0;


#define DATAREADCOMMAND   2
#define DATAREADSETUP 0
#define DATAREADDATA  1


#define ADCSTOP 0x0
#define ADCRUN  0x1


volatile uint8_t adcaction=ADCSTOP;
volatile uint8_t adcstate=ADCSTOP;

void adcStop(void)
{
 adcaction=ADCSTOP;
 while(adcstate==ADCRUN) yield();
}


void adcStart(void)
{
 adcaction=ADCRUN;
 channel=0;
 datareadyevent();
}


void datareadyevent(void)
{
 i2c_txn_t *tn;
 static char       tnbuff[sizeof(i2c_txn_t)+DATAREADYQUEUE*sizeof(i2c_op)];
 static char       buffw[4];
 static char       buffw2[4]; 


 if(adcaction!=adcstate)
 {
  adcstate=adcaction;
  if(adcstate==ADCSTOP) return;
 }

 tn=(i2c_txn_t *)tnbuff;

 i2c_txn_init(tn, DATAREADYQUEUE);

// read result

  // Check for buffer full.
  if((isrBuf!=NULL) && (isrBuf->count >= DATA_DIM16)) 
  {
    // Put buffer isr In full queue.
    uint8_t tmp = fullHead;  // Avoid extra fetch of volatile fullHead.
    fullQueue[tmp] = (block_t*)isrBuf;
    fullHead = queueNext(tmp);

    // Set buffer needed and clear overruns.
    isrBufNeeded = true;
    isrOver = 0;
  }


 // Check for buffer needed.
  if(isrBufNeeded) 
  {
   if(emptyHead==emptyTail) 
   {
    if(isrOver<0XFFFF) 
    {
     isrOver++;
     if(isrBuf!=NULL && isrBuf->count>0) isrBuf->count--;
    }
   }
   else
   {

  // Remove buffer from empty queue.
    isrBuf = emptyQueue[emptyTail];
    emptyTail = queueNext(emptyTail);
    isrBuf->count = 0;
    isrBuf->overrun = isrOver;
    isrBufNeeded = false;

    gps_time(&isrBuf->seconds,&isrBuf->microseconds);
   }
  }
  // Store ADC data.
 
 
 buffw[0]=ADS1015_REG_POINTER_CONVERT;
 i2c_op_init_wr(&tn->ops[DATAREADSETUP],I2C_ADDRESS,(uint8_t*)buffw,1);
 i2c_op_init_rd(&tn->ops[DATAREADDATA],I2C_ADDRESS,(uint8_t*)(isrBuf->data+isrBuf->count++),2);

// set up for next reading

 uint16_t config =  ADS1015_REG_CONFIG_CLAT_NONLAT  | // Non-latching (default val)
                    ADS1015_REG_CONFIG_CPOL_ACTVLOW | // Alert/Rdy active low   (default val)
                    ADS1015_REG_CONFIG_CMODE_TRAD   | // Traditional comparator (default val)
                    ADS1015_REG_CONFIG_DR_6600SPS   | // 1600 samples per second (default)
                    ADS1015_REG_CONFIG_MODE_SINGLE  | 
                    ADS1015_REG_CONFIG_CQUE_1CONV;

 adsGain_t m_gain = GAIN_TWOTHIRDS;


 channel++;
 if(channel>2) channel=0;

 config |= m_gain;

  // Set single-ended input channel
  switch (channel)
  {
    case (0):
      config |= ADS1015_REG_CONFIG_MUX_SINGLE_0;
      break;
    case (1):
      config |= ADS1015_REG_CONFIG_MUX_SINGLE_1;
      break;
    case (2):
      config |= ADS1015_REG_CONFIG_MUX_SINGLE_2;
      break;
    case (3):
      config |= ADS1015_REG_CONFIG_MUX_SINGLE_3;
      break;
  }

  // Set 'start single-conversion' bit
  config |= ADS1015_REG_CONFIG_OS_SINGLE;



 buffw2[0]=ADS1015_REG_POINTER_CONFIG;
 buffw2[1]=config>>8;
 buffw2[2]=config & 0xFF;

 i2c_op_init_wr(&tn->ops[DATAREADCOMMAND],I2C_ADDRESS,(uint8_t*)buffw2,3);

 i2c_post(tn);
}


//**********************************************************************************************************************************




// PPS interrupt

unsigned long rtc_seconds=0;
unsigned long rtc_offset=0;
unsigned long rtc_real=1000000;
unsigned long rtc_actual=1000000;


void rtc_tick(void)
{
 unsigned long nowtime=micros();
 unsigned long deltatime=nowtime-rtc_offset;
 unsigned int seconds=(deltatime+499999)/1000000;

 rtc_real=seconds*1000000;
 rtc_actual=deltatime;
 rtc_seconds+=seconds;
 rtc_offset=nowtime;
}


unsigned long rtc_micros(void)
{
 unsigned long deltatime=micros()-rtc_offset;

 deltatime=(((double)deltatime)*((double)rtc_real)/((double)rtc_actual));

 return(rtc_seconds*1000000+deltatime);
}


ISR (PCINT2_vect)
{
 // handle pin change interrupt for D0 to D7 here
 if (PIND & bit (4)) rtc_tick();
}  // end of PCINT2_vect





unsigned long gps_seconds=0;
unsigned long gps_offset=0;
unsigned long gps_real=1000000;
unsigned long gps_actual=1000000;



void gps_tick(void)
{
 unsigned long nowtime=rtc_micros();
 unsigned long deltatime=nowtime-gps_offset;
 unsigned int seconds=(deltatime+499999)/1000000;

 gps_real=seconds*1000000;
 gps_actual=deltatime;
 gps_seconds+=seconds;
 gps_offset=nowtime;
}



void gps_time(unsigned long * seconds,unsigned long * microseconds)
{
 unsigned long deltatime=rtc_micros()-gps_offset;

 deltatime=(((double)deltatime)*((double)gps_real)/((double)gps_actual));

 unsigned int deltaseconds=deltatime/1000000;

 *seconds=gps_seconds+deltaseconds;
 *microseconds=deltatime-deltaseconds*1000000;
}




void setup() 
{
 Serial.begin(9600);

 if (ERROR_LED_PIN >= 0) 
 {
  pinMode(ERROR_LED_PIN, OUTPUT);
 }

 Serial.print(F("FreeStack: "));
 Serial.println(FreeStack());

 // initialize file system.
 if(!sd.begin(SD_CS_PIN, SPI_FULL_SPEED)) 
 {
  sd.initErrorPrint();
  fatalBlink();
 }
  
 i2c_init();
 i2c_writeregister(I2C_ADDRESS, ADS1015_REG_POINTER_HITHRESH, 0x8000);
 i2c_writeregister(I2C_ADDRESS, ADS1015_REG_POINTER_LOWTHRESH, 0x0);

 pinMode(2, INPUT);
 attachInterrupt(digitalPinToInterrupt(2), datareadyevent, FALLING);

 pinMode(3, INPUT);
 attachInterrupt(digitalPinToInterrupt(3), gps_tick, FALLING);

 PCMSK2 |= bit (PCINT20);  // want pin 4
 PCIFR  |= bit (PCIF2);    // clear any outstanding interrupts
 PCICR  |= bit (PCIE2);    // enable pin change interrupts for D0 to D7
 pinMode (4, INPUT);

 setupgps();
}

void loop() 
{
 do 
 {
  delay(10);
 } while (Serial.available() && Serial.read() >= 0);

  Serial.println();
  Serial.println(F("type:"));
  Serial.println(F("r - record ADC data"));

 while(!Serial.available()) 
 {
  yield();
  loopgps();
 }

 char c = tolower(Serial.read());
 if (ERROR_LED_PIN >= 0) 
 {
  digitalWrite(ERROR_LED_PIN, LOW);
 }
 
 do 
 {
  delay(10);
 } while (Serial.available() && Serial.read() >= 0);

 if (c == 'r') logData();
}


